/**********************/
/* Logging facilities */


var lWarn = { logln: function (o) {} };
var lInfo = { logln: function (o) {} };

var LOGFILE_NAME = "hue-controller.log";
lWarn = new Logger(LOGFILE_NAME);
//lInfo = new Logger(LOGFILE_NAME); /* NOTE: this has to be commented out for release builds to avoid saturating the system with logging activity */

/* NOTE: this is enabled for developer builds, to help with prifiling. Comment out for release builds. */
var lInfoOrig = lInfo;
lInfo =
{
	timebase: (new Date().getTime()),

	logln: function (o)
	{
		// Print milliseconds since this script was loaded
		var ts = String(new Date().getTime() - this.timebase);
		lInfoOrig.logln(ts + ': ' + o);
	}
};
lInfo.logln("---");




/******************/
/* Property names */


var hueBridgeNodeName = "hueBridge";
var hueBridgeIDPropName = hueBridgeNodeName + "/id";
var hueBridgeAddrPropName = hueBridgeNodeName + "/ipaddress";
var hueBridgeAddrManualPropName = hueBridgeNodeName + "/manual";
var hueRegisterStatusPropName = hueBridgeNodeName + "/registerStatus";
var hueRegisterErrorMsgPropName = hueBridgeNodeName + "/registerErrorMsg";
var hueBridgeApiTokenName = hueBridgeNodeName + '/bridgeApiToken';
var hueBridgeZoneAssocPropName = hueBridgeNodeName + "/lightZoneAssociations";
var hueBridgeZoneUseCountPropName = hueBridgeNodeName + "/zoneUseCount";
var hueBridgeContactedBeforePropName = hueBridgeNodeName + "/contactedBefore";
var hueBlinkingIDPropName = hueBridgeNodeName + "/hueBlinkingID";
var hueBlinkingAttributesPropName = hueBridgeNodeName + "/hueBlinkingAttributes";



/*****************/
/* Miscellaneous */


var blinkingTimeout = 3; // in seconds
var defaultLightTransitionTime = 4; // in 100ms steps, example: 4 = 400ms
var dimmingStepSize = 20; // how much the brightness of lamps is incremented/decremented for each dimming call


var HueBridgeRegisterStatus =
{
	OK : 0,
	NEEDS_LINK_BUTTON_PRESSED : 1,
	ERROR_OTHER : 2,
	ERROR_MISSING_INFO : 3,
	ERROR_NO_HUE_BRIDGE_FOUND : 4
};


function htmlUnescape(str)
{
	if (typeof(str) === "string")
		return str.replace(/&amp;/g, '&').replace(/&quot;/g, '"').replace(/&lt;/g, '<').replace(/&gt;/g, '>').replace(/&#039;/g, '\'');
	else
		return str;
}


function getExceptionObjectDetails(err)
{
	var name = err.name;
	name = (name === undefined) ? null : String(name);

	var msg = err.message;
	msg = (msg === undefined) ? null : String(msg);

	var details = "";
	if ((name === "") && (msg === null))
		details = "<unknown error>";
	if ((name === "") && (msg !== null))
		details = msg;
	else if ((name !== "") && (msg === null))
		details = name;
	else
		details = name + ": " + msg;

	var fileName = (err.fileName === undefined) ? null : String(err.fileName);
	var lineNumber = (err.lineNumber === undefined) ? null : String(err.lineNumber);

	if (fileName !== null)
	{
		details += " (" + fileName;
		if (lineNumber != null)
			details += ":" + lineNumber;
		details += ")";
	}

	return details;
}




/******************/
/* CURL functions */


/* Issue HTTP GET request to a given URL using cURL
 *
 * @param url full URL
 * @return result of query, or false if there was an error */
function curlGET(url)
{
	lInfo.logln("curlGET: query from " + url);

	var c = new easycurl;
	c.setopt(c.CURLOPT_URL, url);
	c.setopt(c.CURLOPT_TIMEOUT, 4);
	c.setopt(c.CURLOPT_CONNECTTIMEOUT, 3);

	try
	{
		var resultText = c.perform();
		lInfo.logln("curlGET: returns: " + resultText);
		return resultText;
	}
	catch (err)
	{
		lWarn.logln("curlGET: cURL error: " + err.message);
		return false;
	}
}


/* Issue HTTP PUT or POST request to a given URL using cURL
 *
 * @param url full URL
 * @param data string to be sent in PUT
 * @param method optional string specifying the HTTP method to use (if not given, the default method is PUT)
 * @return result of query, or false if there was an error */
function curlSEND(url, data, method)
{
	if (typeof(method) === 'undefined')
		method = 'PUT';

	lInfo.logln("curlSEND: send request to " + url + ": " + data + " using method " + String(method));

	var c = new easycurl;
	c.setopt(c.CURLOPT_URL, url);
	c.setopt(c.CURLOPT_TIMEOUT, 4);
	c.setopt(c.CURLOPT_CONNECTTIMEOUT, 3);
	c.setopt(c.CURLOPT_POSTFIELDSIZE, data.length);
	c.setopt(c.CURLOPT_COPYPOSTFIELDS, data);

	if (method == 'POST')
		c.setopt(c.CURLOPT_POST, 1); // use POST
	else
		c.setopt(c.CURLOPT_CUSTOMREQUEST, method); // use custom method

	try
	{
		var resultText = c.perform();
		lInfo.logln("curlSEND: result: " + resultText);
		return resultText;
	}
	catch (err)
	{
		lWarn.logln("curlSEND: cURL error: " + err.message);
		return false;
	}
}




/************************/
/* HUE bridge functions */


/* finds IDs and IP addresses of all the bridges using the Philips Portal API
 *
 * @return array with IPs of all the discovered bridges */
function findHueBridges()
{
	var response = curlGET('http://www.meethue.com/api/nupnp');
	response = JSON.parse(unescape(response));
	if (response === false)
		return [];
	else
		return response;
}


/* updates the HUE bridge address and ID values in the property tree
 * either by autodetection using findHueBridges() or by manually setting an IP address
 *
 * @param hueBridgeIPAddress Optional parameter to specify an IP address, as a string
 *        If the parameter is not specified, or an empty string is given, autodetection is used
 */
function updateHueBridgeAddress(hueBridgeIPAddress)
{
	var bridgeID = null;
	var bridgeIPAddress = null;
	var bridgeAddressSetManually = false;

	/* First, check if an IP address is already stored in the property tree.
	 * If this is the case, don't do anything. If an IP address was manually specified,
	 * also don't do anything unless the manually given IP address is different from
	 * the one stored in the property tree
	 * (Otherwise it would not be possible to configure a new IP address) */
	bridgeIPAddress = htmlUnescape(Property.getProperty(hueBridgeAddrPropName));
	if (bridgeIPAddress && (bridgeIPAddress == hueBridgeIPAddress))
	{
		lInfo.logln("updateHueBridgeAddress: HUE bridge information already queried");
		return;
	}

	if (hueBridgeIPAddress)
	{
		lInfo.logln("updateHueBridgeAddress: HUE bridge IP address set manually: " + JSON.stringify(hueBridgeIPAddress));

		bridgeIPAddress = hueBridgeIPAddress;
		bridgeID = "< bridge IP address manually specified >";
		bridgeAddressSetManually = true;
	}
	else
	{
		var hueBridges = findHueBridges();
		if (hueBridges.length == 0)
		{
			lWarn.logln("updateHueBridgeAddress: no HUE bridges found");

			/* no HUE bridge found -> make sure no address is set by deleting
			 * the nodes from the property tree */
			var baseNode = Property.getNode("/scripts/hue-controller");
			baseNode.removeChild(hueBridgeNodeName);
			Property.store();

			/* Make sure the code below does not write anything into the property tree */
			bridgeIPAddress = null;
			bridgeID = null;
		}
		else
		{
			// Use information about the first discovered bridge
			var hueBridgeInfo = hueBridges[0];
			bridgeID = hueBridgeInfo.id;
			bridgeIPAddress = hueBridgeInfo.internalipaddress;
			bridgeAddressSetManually = false;

			lInfo.logln("updateHueBridgeAddress: HUE bridge found: " + JSON.stringify(hueBridgeInfo));
		}
	}

	/* If either a manual IP address is used, or a bridge was discovered, write the values to
	 * the property tree and mark them as persistent */
	if (bridgeIPAddress)
	{
		Property.setProperty(hueBridgeIDPropName, bridgeID);
		Property.setProperty(hueBridgeAddrPropName, bridgeIPAddress);
		Property.setProperty(hueBridgeAddrManualPropName, bridgeAddressSetManually);

		Property.setFlag(hueBridgeAddrManualPropName, 'ARCHIVE', true);
		Property.setFlag(hueBridgeIDPropName, 'ARCHIVE', bridgeAddressSetManually);
		Property.setFlag(hueBridgeAddrPropName, 'ARCHIVE', bridgeAddressSetManually);

		Property.store();
	}
}


/* Returns an object containing the stored ID and the IP address of the configured bridge
 * If nothing is configured, returns null */
function hueBridgeBaseInfo()
{
	var bridgeID = htmlUnescape(Property.getProperty(hueBridgeIDPropName));
	var bridgeAddr = htmlUnescape(Property.getProperty(hueBridgeAddrPropName));

	return (!bridgeID || !bridgeAddr) ? null : { id: bridgeID, address: bridgeAddr };
}


/* Constructs the base URL for HUE bridge communication
 * The base URL has the format:  http://<IP address>/api
 *
 * @return: the base URL, or false if no HUE bridge IP address has been stored */
function hueBridgeBaseURL()
{
	var bridgeAddr = htmlUnescape(Property.getProperty(hueBridgeAddrPropName));
	if (!bridgeAddr)
	{
		lWarn.logln("hueBridgeBaseURL: no bridge URL configured");
		return false;
	}
	return "http://" + bridgeAddr + '/api';
}


/* Constructs the access URL for HUE bridge communication
 * For most cases, this is the URL to use (instead of the base URL).
 * It appends the API token to the URL. The API token is the username used for accessing
 * HUE bridge functionality, and is defined in the HUE bridge registration process.
 * The acess URL has the format:  <base URL>/<API token>
 *
 * @return the access URL, or false if no HUE bridge IP address or no API token has been stored */
function hueBridgeAccessURL()
{
	var baseURL = hueBridgeBaseURL();
	if (!baseURL)
		return false;

	// check if there is an API token
	var apiToken = htmlUnescape(Property.getProperty(hueBridgeApiTokenName));
	if (!apiToken || (apiToken.length == 0))
	{
		// No token exists - registration is necessary
		lInfo.logln("hueBridgeAccessURL: no API token present, must perform registration first");
		return false;
	}

	return baseURL + '/' + apiToken;
}


/* Check the result of a HUE query.
 * If there was an authorization error, the register status is reset to NEEDS_LINK_BUTTON_PRESSED,
 * to inform the rest of the system that the user needs to press the link button before proceeding.
 *
 * @param hueResult Result to check
 * @return true if the result is OK and authorization is OK, false otherwise */
function hueCheckResult(hueResult)
{
	lInfo.logln("hueCheckResult: " + JSON.stringify(hueResult));

	// Iterate through all result objects, check if all are OK.
	// Also tolerate query results in other formats.
	for (var ei = 0; ei < hueResult.length; ++ei)
	{
		var res = hueResult[ei];
		if (res)
		{
			var err = res.error;
			if (err)
			{
				// investigate error if it is a known one
				if ((err.type == 1) && (err.description == 'unauthorized user'))
				{
					lInfo.logln("hueCheckResult: Authorization error -> need to obtain new application token");

					// Clear invalid API token, and set the register state to signal that the link button needs to be pressed
					htmlUnescape(Property.getProperty(hueBridgeApiTokenName));
					Property.setProperty(hueRegisterStatusPropName, HueBridgeRegisterStatus.NEEDS_LINK_BUTTON_PRESSED);
					Property.store();

					return false;
				}	
			}			
		}
	}

	return true;
}


/* Convenience wrapper around hueCheckResult() which parses a JSON string
 * and returns false if either the JSON string is empty or invalid, or if
 * the parsed array of result objects is determined to be invalid by hueCheckResult().
 *
 * @return true if the JSON-encoded result is OK, false otherwise */
function hueCheckResultText(hueResultText)
{
	if (hueResultText && (hueResultText.length > 0))
	{
		var hueResult = null;

		try
		{
			hueResult = JSON.parse(hueResultText);
		}
		catch (err)
		{
			lWarn.logln("hueCheckResultText: Error while parsing the JSON encoded result string " + String(hueResultText) + ": " + String(err));
			return false;
		}

		return hueCheckResult(hueResult);
	}
	else
	{
		// String is invalid
		return false;
	}
}


/* Issue a query to the HUE bridge using HTTP GET
 *
 * @param queryAddress query address relative to the access URL
 * @return Response object, or null if there was an error */
function hueQuery(queryAddress)
{
	var url = hueBridgeAccessURL();
	lInfo.logln("hueQuery: using access URL " + String(url) + " for query address " + String(queryAddress));

	var result = null;

	if (url!==false)
	{
		var resultText = curlGET(url + "/" + queryAddress);
		if (resultText !== false)
		{
			result = JSON.parse(resultText);
			if (result)
			{
				// got JSON result - check for basic error
				if (!hueCheckResult(result))
					result = null;
				// otherwise, pass on actual result
			}
		}
	}

	return result;
}


function dummyCURLCallback(foo)
{
	return true;
}


/* Send a command to the HUE bridge using a HTTP method (typically PUT or POST)
 *
 * @param cmdAddress command address relative to the access URL
 * @param cmd command object to send to the HUE bridge (will automatically be JSON-encoded prior to sending)
 * @param method string specifying the HTTP method to use (typically PUT or POST)
 * @param doAsynchronous if true, the HTTP request can be done asynchronously, and the caller does not care about the response
 * @return true if the command was executed successfully, false otherwise (or if an error ocurred)
 */
function hueCommand(cmdAddress, cmd, method, doAsynchronous)
{
	var url = hueBridgeAccessURL();
	var cmdJSON = JSON.stringify(cmd);
	lInfo.logln("hueCommand: using access URL " + String(url) + " for command address " + String(cmdAddress) + " command " + cmdJSON + " and method " + String(method));

	var url = hueBridgeAccessURL() + "/" + cmdAddress;
	if (url !== false)
	{
		if (doAsynchronous)
		{
			var easy = new easycurl();

			switch (method)
			{
				case 'GET':
					easy.setopt(easy.CURLOPT_HTTPGET, 1);
					break;
				case 'POST':
					easy.setopt(easy.CURLOPT_POST, 1);
					break;
				default:
					easy.setopt(easy.CURLOPT_CUSTOMREQUEST, method || 'PUT');
			}

			easy.setopt(easy.CURLOPT_URL, url);
			easy.setopt(easy.CURLOPT_TIMEOUT, 4);
			easy.setopt(easy.CURLOPT_CONNECTTIMEOUT, 3);
			easy.setopt(easy.CURLOPT_POSTFIELDSIZE, cmdJSON.length);
			easy.setopt(easy.CURLOPT_COPYPOSTFIELDS, cmdJSON);

			easy.asyncdone = dummyCURLCallback;
			easy.perform_async(dummyCURLCallback);

			return true;
		}
		else
		{
			var resultText = curlSEND(url, cmdJSON, method);
			if (resultText)
			{
				// return true or false depending on returned status
				return hueCheckResultText(resultText);
			}
		}
	}

	// if we get here, command has failed anyway
	return false;
}


/* Register this application in the HUE bridge
 * If it is registered already, this function does nothing.
 * If the user needs to press the link button, the register status property is set to NEEDS_LINK_BUTTON_PRESSED . */
function hueBridgeRegisterApp()
{
	var apitoken = false;
	var regval = HueBridgeRegisterStatus.OK;
	var regErrorText = null;

	/* before registering, verify we are not yet successfully registered already
	 * by trying to query the discovered lights */
	var cfg = hueQuery('lights');
	if (cfg !== null)
	{
		lInfo.logln("hueBridgeRegisterApp: already registered, lights: " + JSON.stringify(cfg));

		// since we are registered already, mark the register status as OK
		Property.setProperty(hueRegisterStatusPropName, HueBridgeRegisterStatus.OK);
		Property.store();

		return;
	}

	// can't query the discovered lights, so we might need to register
	var url = hueBridgeBaseURL();
	lInfo.logln("hueBridgeRegisterApp: base url: " + url);

	if (url !== false)
	{
		/* we have an URL, so try register there
		 * -> base user name on the MAC address of eth0, should be unique */
		var macaddr = htmlUnescape(Property.getProperty("/system/host/interfaces/eth0/mac"));

		// username must have at least 10 characters; "dSS" + MAC address fulfill this requirement
		var username = 'dSS';
		for (var i = 0; i < macaddr.length; i += 3)
			username += macaddr.substr(i, 2);

		lInfo.logln("hueBridgeRegisterApp: trying registration with user name: " + username);

		/* first just try with this username, we might already be registered
		 * (using curlGET() directly instead of hueQuery(), since the latter uses
		 * hueBridgeAccessURL(), which fails prior to registration) */
		var cfg = curlGET(url + username + '/lights');

		if (hueCheckResultText(cfg))
		{
			// username is ok as API token
			lInfo.logln("hueBridgeRegisterApp: username is ok as API token, no re-registering needed: " + username);
			apitoken = username;
			regval = HueBridgeRegisterStatus.OK;
		}
		else
		{
			// need real registration
			var regData =
			{
				"username": username,
				"devicetype": "plan44.ch huecontrol digitalSTROM add-on"
			};

			var regDataText = JSON.stringify(regData);
			var regResultText = curlSEND(url, regDataText, 'POST');
			lInfo.logln("hueBridgeRegisterApp: bridge returns: " + regResultText);

			/* HUE bridge might suggest a new username as API token,
			 * so parse the result for an username */
			if (regResultText)
			{
				regval = HueBridgeRegisterStatus.ERROR_OTHER;

				var regResult = null;

				try
				{
					regResult = JSON.parse(regResultText);
				}
				catch (err)
				{
					lWarn.logln('hueBridgeRegisterApp: error while parsin JSON result: ' + String(err));
				}

				if (regResult)
				{
					var statusObj = regResult[0];
					if (statusObj)
					{
						var successObj = statusObj.success;
						if (successObj)
						{
							apitoken = successObj.username;
							lInfo.logln("hueBridgeRegisterApp: registered successfully, new API token is: " + apitoken);
							regval = HueBridgeRegisterStatus.OK;
						}
						else
						{
							lInfo.logln("hueBridgeRegisterApp: registration failed, user needs to press link button");
							regval = HueBridgeRegisterStatus.NEEDS_LINK_BUTTON_PRESSED;
						}
					}
				}

				if (regval === HueBridgeRegisterStatus.ERROR_OTHER)
					regErrorText = 'Invalid registration result: ' + regResultText;
			}
			else
				regval = HueBridgeRegisterStatus.ERROR_NO_HUE_BRIDGE_FOUND;
		}
	}
	else
	{
		// no bridge present
		regval = HueBridgeRegisterStatus.ERROR_NO_HUE_BRIDGE_FOUND;
	}

	// save new API token if we have one
	if (apitoken)
	{
		Property.setProperty(hueBridgeApiTokenName, apitoken);
		Property.setFlag(hueBridgeApiTokenName, 'ARCHIVE', true);
		Property.store();
		lInfo.logln("hueBridgeRegisterApp: saved new API token to tree");
	}
	else
	{
		Property.setProperty(hueBridgeApiTokenName, '');
		Property.setFlag(hueBridgeApiTokenName, 'ARCHIVE', true);
		Property.store();
		lWarn.logln("hueBridgeRegisterApp: not registered, no API token");
	}

	/* store new register status
	 * this must not be persistent, because potentially stale status values
	 * (for example after a reset) could cause undefined behavior */ 
	Property.setProperty(hueRegisterStatusPropName, regval);
	if ((regval === HueBridgeRegisterStatus.ERROR_OTHER) && regErrorText)
		Property.setProperty(hueRegisterErrorMsgPropName, regErrorText);
	Property.setFlag(hueRegisterStatusPropName, 'ARCHIVE', false);
	Property.store();
}


/* Queries the HUE bridge for a list of connected and autodetected HUE lights.
 *
 * @return array of objects describing the ID and name of the detected lights */
function hueLights()
{
	var lights = hueQuery('lights');
	var lightsObjs = [];

	lInfo.logln("hueLights: bridge returns: " + JSON.stringify(lights));

	if (lights)
	{
		/* HUE bridge returned a non-empty object containing information about lights.
		 * For further use, this has to be converted to an array of light objects,
		 * each object containing an ID (= the property name in the original light object)
		 * and the name (= the value in the original object).
		 *
		 * Example: the HUE bridge returns: { '1': 'light1', '2': 'light2' }
		 * this loop converts it to: [ { lightID: '1', name: 'light1' }, { lightID: '2', name: 'light2' } ]
		 */
		for (var light in lights)
		{
			lightsObjs.push({ lightID: light, name: lights[light].name });
		}
		return lightsObjs;
	}
	else
	{
		/* No light information received from the HUE bridge */

		if (htmlUnescape(Property.getProperty(hueRegisterStatusPropName)) == HueBridgeRegisterStatus.OK)
		{
			// return dummy light IDs for testing and development
			// XXX enable these for testing and development only
			return [
			/*	{ "lightID": "1", "name" : "test lamp 1" },
				{ "lightID": "2", "name" : "test lamp 2" },
				{ "lightID": "3", "name" : "test lamp 3" },
				{ "lightID": "4", "name" : "test lamp 4" },
				{ "lightID": "5", "name" : "test lamp 5" }*/
			];
		}
		else
			return [];
	}

}


function getHueLightZoneAssotiations()
{
	var retval = [];
	var zonesNode = Property.getNode(hueBridgeZoneAssocPropName);
	if (zonesNode !== null)
	{
		var childNodes = zonesNode.getChildren();
		for (var i = 0; i < childNodes.length; ++i)
		{
			var childNode = childNodes[i];
			retval.push({
				lightID: childNode.getName(),
				zoneID: childNode.getValue()
			});
		}
	}

	return retval;
}


/* scene IDs and their meanings according to dss-types.js and dss-datamodel.js :
 *
 *   scene ID    meaning
 *   0           off
 *   32          scene 10
 *   34          scene 20
 *   36          scene 30
 *   38          scene 40
 *   22          scene 14
 *   67          standby
 *   68          deep off
 *   69          sleeping
 *   40          fade off
 *
 * With scene IDs < 64, triggers for group 0 and group 1 are created;
 * with scene IDs >= 64, only a trigger for group 0 is created
 */
var offSceneIDs = [ 0, 32, 34, 36, 38, 22, 67, 68, 69, 40 ];

function isOffSceneID(sceneID)
{
	return offSceneIDs.indexOf(parseInt(sceneID)) != -1;
}


function isGlobalOffSceneID(sceneID)
{
	switch (sceneID)
	{
		case 0:
		case 67:
		case 68:
			return true;
		default:
			return false;
	}
}


function setHueLightZoneAssociation(lightID, zoneID)
{
	if ((zoneID !== null) && (zoneID !== undefined) && (zoneID !== ''))
	{
		// First, check if any lights are already associated with the zone
		var existingLights = getLightsInZone(zoneID);

		// Now associate this light with the zone
		lInfo.logln('setHueLightZoneAssociation: associating light of ID ' + String(lightID) + ' to zone of ID ' + String(zoneID));
		var path = hueBridgeZoneAssocPropName + "/" + String(lightID);
		var oldZoneID = Property.getProperty(path);
		Property.setProperty(path, zoneID);
		Property.setFlag(path, 'ARCHIVE', true);

		lInfo.logln("setHueLightZoneAssociation: old zone ID: " + String(oldZoneID) + " new zone ID: " + String(zoneID));

		if (zoneID !== oldZoneID)
		{
			var oldZoneUseCount = (oldZoneID !== null) ? Property.getProperty(hueBridgeZoneUseCountPropName + "/" + String(oldZoneID)) : null;
			var newZoneUseCount = Property.getProperty(hueBridgeZoneUseCountPropName + "/" + String(zoneID));

			lInfo.logln("setHueLightZoneAssociation: count of old zone ID: " + String(oldZoneUseCount) + " new zone ID: " + String(newZoneUseCount));

			if (oldZoneUseCount !== null)
			{
				oldZoneUseCount--;
				if (oldZoneUseCount <= 0)
				{
					lInfo.logln("setHueLightZoneAssociation: decremented old zone ID count to 0 -> disconnecting light device");
					Property.getNode(hueBridgeZoneUseCountPropName).removeChild(oldZoneID);
					try {
					getZoneByID(oldZoneID).removeConnectedDevice(1);
					}
					catch (e) {}
				}
				else
				{
					Property.setProperty(hueBridgeZoneUseCountPropName + "/" + String(oldZoneID), oldZoneUseCount);
					Property.setFlag(hueBridgeZoneUseCountPropName + "/" + String(oldZoneID), 'ARCHIVE', true);
				}
			}

			if (newZoneUseCount !== null)
			{
				newZoneUseCount++;
				Property.setProperty(hueBridgeZoneUseCountPropName + "/" + String(zoneID), newZoneUseCount);
			}
			else
			{
				lInfo.logln("setHueLightZoneAssociation: incremented new zone ID count for the first time -> connecting light device");
				Property.setProperty(hueBridgeZoneUseCountPropName + "/" + String(zoneID), 1);
				try {
					getZoneByID(zoneID).addConnectedDevice(1);
				} 
				catch(e) {
					
				}
				
			}
			Property.setFlag(hueBridgeZoneUseCountPropName + "/" + String(zoneID), 'ARCHIVE', true);
		}

		Property.store();

		// Visit all existing presets and add light settings for this new light if applicable
		checkLightsOfActivities(ActivityTypes.PRESET);

		// If no other lights were associated, create standard presets
		// for the zone
		if (existingLights.length === 0)
		{
			lInfo.logln('setHueLightZoneAssociation: no lights were associated with zone ' + String(zoneID) + ' -> adding standard presets for this zone');
			createStandardPresets(zoneID);
		}
		else
			lInfo.logln('setHueLightZoneAssociation: zone ' + String(zoneID) + ' already had associated lights - no need to add standard presets');
	}
	else
	{
		lInfo.logln('setHueLightZoneAssociation: removing any association of a zone with light of ID ' + String(lightID));
		var node = Property.getNode(hueBridgeZoneAssocPropName);
		if (node)
			node.removeChild(String(lightID));
	}

/*
	var sceneIDs = null;

	// Create triggers for dimming in this zone
	// scene IDs according to dss-types.js: decrement, increment, min, max
	sceneIDs = [11, 12, 13, 14];
	for (var i = 0; i < sceneIDs.length; ++i)
	{
		var propPath = "preset/predefined/dimming/triggers/" + String(zoneID) + "_" + String(sceneIDs[i]);
		storeTriggerProperty(propPath, "type", "bus-zone-scene");
		storeTriggerProperty(propPath, "zone", zoneID);
		storeTriggerProperty(propPath, "group", 1);
		storeTriggerProperty(propPath, "scene", sceneIDs[i]);
		storeTriggerProperty(propPath, "dsid", -1);
	}
*/

	// Associate this zone with off/standby scene IDs
	for (var i = 0; i < offSceneIDs.length; ++i)
	{
		var sceneID = offSceneIDs[i];
		var numGroups = (sceneID < 64) ? 2 : 1;
		for (var group = 0; group < numGroups; ++group)
		{
			var propPath = "preset/predefined/off/triggers/" + String(zoneID) + "_" + String(sceneID) + "_" + String(group);
			storeTriggerProperty(propPath, "type", "bus-zone-scene");
			storeTriggerProperty(propPath, "zone", zoneID);
			storeTriggerProperty(propPath, "group", group);
			storeTriggerProperty(propPath, "scene", sceneID);
			storeTriggerProperty(propPath, "dsid", -1);
		}
	}

	Property.store();
}


function getLightsInZone(zoneID)
{
	var retval = [];
	var zonesNode = Property.getNode(hueBridgeZoneAssocPropName);
	if (zonesNode !== null)
	{
		var childNodes = zonesNode.getChildren();
		for (var i = 0; i < childNodes.length; ++i)
		{
			var childNode = childNodes[i];
			var lightID = parseInt(childNode.getName());
			var assocZoneID = childNode.getValue();
			if (assocZoneID === zoneID)
				retval.push(lightID);
		}
	}

	return retval;
}


function getLightsInZoneAsObject(zoneID)
{
	var retval = {};
	var zonesNode = Property.getNode(hueBridgeZoneAssocPropName);
	if (zonesNode !== null)
	{
		var childNodes = zonesNode.getChildren();
		for (var i = 0; i < childNodes.length; ++i)
		{
			var childNode = childNodes[i];
			var lightID = parseInt(childNode.getName());
			var assocZoneID = childNode.getValue();
			if (assocZoneID === zoneID)
				retval[lightID] = true;
		}
	}

	return retval;
}


function hueDimLightsInZone(zoneID, dimValue)
{
	var lights = getLightsInZone(zoneID);
	var newSettings = [];
	lInfo.logln("hueDimLightsInZone: dimming lights with ID " + JSON.stringify(lights) + " by " + String(dimValue));
	for (var i = 0; i < lights.length; ++i)
	{
		var lightID = lights[i];
		var attributes = hueQuery('lights/' + String(lightID));
		if (attributes && attributes.state)
		{
			var newBri = attributes.state.bri + dimValue;
			newBri = (newBri < 4) ? 4 : ((newBri > 255) ? 255 : newBri);

			newSettings.push({
				id: lightID,
				color: [
					attributes.state.hue / 65535.0,
					attributes.state.sat / 255.0,
					newBri / 255.0
				],
				on: attributes.state.on,
				blinking: false,
				transitiontime: 0
			});
		}
	}

	sendLightsToHueBridge(newSettings);
}


function hueSwitchOffLightsInZone(zoneID, sceneID)
{
	var lights = getLightsInZone(zoneID);
	var newSettings = [];
	lInfo.logln("hueSwitchOffLightsInZone: switching off lights with ID " + JSON.stringify(lights) + " in zone with ID " + String(zoneID));
	for (var i = 0; i < lights.length; ++i)
	{
		var lightID = lights[i];

		// use default transition time for all scenes except scene 40 (fade off);
		// for that scene, use 30 seconds instead
		var transition_time = (sceneID === 40) ? 300 : defaultLightTransitionTime;

		newSettings.push({id: lightID, blinking: false, on: false, transitiontime: transition_time});
	}

	sendLightsToHueBridge(newSettings);

	return newSettings;
}


/* Returns an object containing HUE bridge information (base information and detected lights).
 *
 * @return: object containing two fields, baseInfo and lights, or {} if an error occurs */
function getHueBridgeInfo()
{
	var info = {};

	try
	{
		{
			lInfo.logln('getHueBridgeInfo: checking if a HUE bridge was found earlier');
			var hueBridgeNode = Property.getNode(hueBridgeNodeName);
			var regStatus = Property.getProperty(hueRegisterStatusPropName);
			if ((hueBridgeNode === null) || (regStatus === HueBridgeRegisterStatus.ERROR_NO_HUE_BRIDGE_FOUND))
			{
				lWarn.logln('getHueBridgeInfo: no HUE bridge was found');

				if (regStatus === null)
				{
					Property.setProperty(hueRegisterStatusPropName, HueBridgeRegisterStatus.ERROR_NO_HUE_BRIDGE_FOUND);
					Property.store();
				}

				return {};
			}
			else if (regStatus != HueBridgeRegisterStatus.OK)
			{
				lWarn.logln('getHueBridgeInfo: HUE bridge error reported earlier');
				return {};
			}
			else
				lInfo.logln('getHueBridgeInfo: HUE bridge was found');
		}

		var baseInfo = hueBridgeBaseInfo();
		var lights = hueLights();

		// trigger a search for lights in the background
		hueFindNewLights();

		if (baseInfo)
		{
			lInfo.logln('getHueBridgeInfo: base info: ' + JSON.stringify(baseInfo));
			info.baseInfo = baseInfo;
		}
		else
		{
			Property.setProperty(hueRegisterStatusPropName, HueBridgeRegisterStatus.ERROR_MISSING_INFO);
			Property.store();

			return {};
		}

		if (lights)
		{
			lInfo.logln('getHueBridgeInfo: lights: ' + JSON.stringify(lights));

			for (var i = 0; i < lights.length; ++i)
			{
				var light = lights[i];
				var assocZone = Property.getProperty(hueBridgeZoneAssocPropName + "/" + String(light.lightID));
				if (assocZone)
					light.zoneID = assocZone;
			}

			info.lights = lights;
		}
		else
		{
			Property.setProperty(hueRegisterStatusPropName, HueBridgeRegisterStatus.ERROR_MISSING_INFO);
			Property.store();

			return {};
		}
	}
	catch (err)
	{
		lWarn.logln("getHueBridgeInfo: " + String(err));

		Property.setProperty(hueRegisterStatusPropName, HueBridgeRegisterStatus.ERROR_OTHER);
		Property.setProperty(hueRegisterErrorMsgPropName, String(err));
		Property.store();

		return {};
	}

	return info;
}


/* Get register status from the property tree.
 *
 * @return the register status; if accessing the status failed, returns ERROR */
function getHueBridgeRegisterStatus()
{
	var registerStatus = htmlUnescape(Property.getProperty(hueRegisterStatusPropName));
	lInfo.logln("getHueBridgeRegisterStatus: status: " + String(registerStatus));
	if ((registerStatus === null) || (registerStatus === undefined))
	{
		registerStatus = HueBridgeRegisterStatus.ERROR_OTHER;
		Property.setProperty(hueRegisterErrorMsgPropName, "missing register status");
	}
	return registerStatus;
}


function hueFindNewLights()
{
	if (hueCommand('lights', {}, 'POST'))
		lInfo.logln('hueFindNewLights: successfully initiated search for new lights');
	else
		lWarn.logln('hueFindNewLights: could not initiate search for new lights');
}




/*********************/
/* Activity handling */


var ActivityTypes =
{
	PRESET: "preset",
	GLOBAL_ACTIVITY: "globalActivity"
};


function checkActivityType(activityType)
{
	for (var validType in ActivityTypes)
	{
		if (ActivityTypes[validType] === activityType)
			return true;
	}

	lWarn.logln("checkActivityType: found invalid activity " + String(activityType));

	return false;
}


function handleBlinkingTimeout(lights)
{
	var retainedLights = [];
	var otherLights = [];

	for (var i = 0; i < lights.length; ++i)
	{
		var light = lights[i];

		if (light.retainOutput)
			retainedLights.push(light);
		else
			otherLights.push(light);
	}

	if (retainedLights.length > 0)
		sendLightsToHueBridge(retainedLights);
	if (otherLights.length > 0)
		stopLightsBlinking(otherLights);
}


function stopLightsBlinking(lights)
{
	for (var i = 0; i < lights.length; ++i)
	{
		var light = lights[i];

		var cmd = { alert: 'none' };
		hueCommand('lights/' + light.id + '/state', cmd, 'PUT', false);
	}
}


/* Sends light properties to the HUE bridge as a command using HTTP PUT.
 *
 * @param lights Array of light structures describing how the affected HUE lights shall behave */
function sendLightsToHueBridge(lights)
{
	var blinkingLights = [];

	for (var i = 0; i < lights.length; ++i)
	{
		var light = lights[i];

		if (light.blinking)
		{
			var curLightAttr = getHueBridgeLightAttributes(light.id);
			curLightAttr.blinking = false; // this prevents endless loops caused by blinking
			curLightAttr.retainOutput = light.ignored;
			curLightAttr.transitiontime = defaultLightTransitionTime; // make sure dimming transition speeds are never stored in blinkingLights
			blinkingLights.push(curLightAttr);
		}
	}

	for (var i = 0; i < lights.length; ++i)
	{
		var light = lights[i];

		if (light.ignored && !(light.blinking))
			continue;

		var cmd = {
			on: light.on,
			alert: light.blinking ? 'lselect' : 'none',
			transitiontime: (light.transitiontime === undefined) ? defaultLightTransitionTime : light.transitiontime
		};

		if (light.color !== undefined)
		{
			cmd.hue = Math.floor(light.color[0] * 65535.0);
			cmd.sat = Math.floor(light.color[1] * 255.0);
		}

		if ((light.color !== undefined) || (light.ctbri !== undefined))
			cmd.bri = light.useCT ? light.ctbri : Math.floor(light.color[2] * 255.0);

		if (light.useCT)
			cmd.ct = Math.floor(light.ct);

		hueCommand('lights/' + light.id + '/state', cmd, 'PUT', false);
	}

	if (blinkingLights.length > 0)
	{
		lInfo.logln("sendLightsToHueBridge: at least one light is blinking - starting blink timeout");

		// store current attributes only if no blinking is active at the moment
		var blinkID = Property.getProperty(hueBlinkingIDPropName);
		if ((blinkID === null) || (blinkID === 0))
			Property.setProperty(hueBlinkingAttributesPropName, JSON.stringify(blinkingLights));

		// generate & store a new ID
		blinkID = Math.floor(Math.random() * 100000000) + 1;
		Property.setProperty(hueBlinkingIDPropName, blinkID);

		// raise event
		(new Event('blinkingTimeout', { time: '+' + String(blinkingTimeout), blinkID: blinkID })).raise();
	}
	else
		Property.setProperty(hueBlinkingIDPropName, 0);
}


function getHueBridgeLightAttributes(lightID)
{
	var attributes = hueQuery('lights/' + String(lightID));
	var retval = { id: lightID };
	if (attributes)
	{
		retval.color = [ attributes.state.hue / 65535.0, attributes.state.sat / 255.0, attributes.state.bri / 255.0 ];
		retval.on = attributes.state.on;
		retval.blinking = (attributes.state.alert != 'none');
		retval.useCT = (attributes.state.colormode === 'ct');
		retval.ct = (attributes.state.ct !== undefined) ? attributes.state.ct : 500;
		retval.ctbri = (attributes.state.bri !== undefined) ? attributes.state.bri : 255;
	}

	return retval;
}


function findFreeActivityID(activityType)
{
	var baseNode = Property.getNode(activityType);
	if (baseNode)
	{
		var children = baseNode.getChildren();
		var maxNumIDs = 10000000;
		for (var candidateID = 0; candidateID < maxNumIDs; ++candidateID)
		{
			var idstr = String(candidateID);
			var child = baseNode.getChild(idstr);

			if (!child)
				return idstr;
		}

		lWarn.logln("[FATAL] findFreeActivityID: could not find free activity ID - there are over " + String(maxNumIDs) + " activities?!");
		return null;
	}
	else
		return "0";
}


function storeTriggerProperty(path, propertyName, propertyValue)
{
	if ((propertyValue === null) || (propertyValue === undefined))
	{
		lWarn.logln('storeTriggerProperty: no value defined for trigger property "' + propertyName + '" (path: ' + path + ')');
		return;
	}

	lInfo.logln("storeTriggerProperty: path: " + path + " name: " + propertyName + " value: " + propertyValue);

	var fullPath = path + "/" + propertyName;
	Property.setProperty(fullPath, propertyValue);
	Property.setFlag(fullPath, "ARCHIVE", true);
}


function deleteActivity(activityID, activityType)
{
	if (!checkActivityType(activityType))
		return;

	activityID = String(activityID);

	lInfo.logln('deleteActivity: deleting activity of type "' + activityType + '" with ID ' + activityID);

	var baseNode = Property.getNode(activityType);
	if (baseNode)
	{
		var childNode = baseNode.getChild(activityID);
		if (childNode)
		{
			unregisterTrigger(childNode.getPath(), "hueController.handleTrigger", { });
			baseNode.removeChild(activityID);
			Property.store();
		}
		else
		{
			lWarn.logln('deleteActivity: found no node for activity of type "' + activityType + '" with ID ' + activityID);
		}
	}
	else
	{
		lInfo.logln('getActivity: found no subnode for activities of type "' + activityType + '"');
	}
}


function getZoneSceneInfoForActivity(activityID, activityType)
{
	var basePropertyPath = getBaseActivityPath(activityID, activityType);
	lInfo.logln("setActivity: basePropertyPath: " + basePropertyPath);

	var triggersPath = basePropertyPath + "/triggers";
	var triggersNode = Property.getNode(triggersPath);
	if (triggersNode === null)
	{
		lWarn.logln("getZoneSceneInfoForActivity: could not get triggers node for activity of type " + activityType + " and ID " + activityID);
		return null;
	}

	var children = triggersNode.getChildren();
	if (children === null)
	{
		lWarn.logln("getZoneSceneInfoForActivity: could not get child trigger nodes for activity of type " + activityType + " and ID " + activityID);
		return null;
	}

	for (var i = 0; i < children.length; ++i)
	{
		var child = children[i];

		var triggerType = child.getChild("type").getValue();

		if ((triggerType === 'zone-scene') || (triggerType === 'bus-zone-scene'))
		{
			lInfo.logln("getZoneSceneInfoForActivity: found scene trigger for activity of type " + activityType + " and ID " + activityID);

			var sceneID = child.getChild("scene").getValue();
			var groupID = child.getChild("group").getValue();
			var zoneID = child.getChild("zone").getValue();

			lInfo.logln("getZoneSceneInfoForActivity: scene ID " + String(sceneID) + " group ID " + String(groupID) + " zone ID " + String(zoneID));

			if ((sceneID === null) || (groupID === null) || (zoneID === null))
			{
				lWarn.logln("getZoneSceneInfoForActivity: invalid zone scene data - returning null");
				return null;
			}
			else
				return { sceneID: sceneID, groupID : groupID, zoneID: zoneID };
		}
	}

	return null;
}


function testActivity(activityID, activityType)
{
	var propPath = activityType + "/" + activityID;

	// call scene if activity has a scene based trigger, then send light settings to HUE
	// otherwise, just send the settings
	// (even with callScene, the settings are explictiely sent, because for unknown reasons
	// the handleTrigger handler is not called by the DSS after the callScene event is sent)
	// also, global activities never use callScene
	if (activityType === ActivityTypes.PRESET)
	{
		var zoneSceneInfo = getZoneSceneInfoForActivity(activityID, activityType);
		if (zoneSceneInfo !== null)
		{
			lInfo.logln("testActivity: activity is scene-based => using callScene");

			var z = getZoneByID(zoneSceneInfo.zoneID);
			if (z === null)
			{
				lWarn.logln("testActivity: no zone with ID " + String(zoneID) + " found -> sending light settings directly without callScene");
			}
			else
			{
				z.callScene(zoneSceneInfo.groupID, zoneSceneInfo.sceneID, true);
				return;
			}
		}
		else
			lInfo.logln("testActivity: activity is not scene-based => sending light settings to HUE lamps directly without callScene");
	}
	else
		lInfo.logln("testActivity: activity is a global activity => sending light settings to HUE lamps directly without callScene");

	// This point is reached either because no scene-based trigger or because
	// no zone with the given ID was found

	{

		var lightsJSON = htmlUnescape(Property.getProperty(propPath + "/lights"));

		if (lightsJSON === null)
		{
			lWarn.logln('testActivity: found no lights field for activity at path "' + propPath + '"');
			return;
		}

		var lights = null;
		try
		{
			lights = JSON.parse(lightsJSON);
			lInfo.logln("testActivity: sending light properties to HUE bridge: " + lightsJSON);
			sendLightsToHueBridge(lights);
		}
		catch (err)
		{
			lWarn.logln('testActivity: error while parsing JSON data in lights field for activity at path "' + propPath + '": ' + err);
			return;
		}
	}
}


function testLight(params)
{
	lInfo.logln("testLight: setting test values for light with ID " + String(params.lightID) + ", params: " + JSON.stringify(params));

	var lights =
	[
		{
			id: params.lightID,
			on: params.on,
			color: params.color ? params.color : [ 1, 0, 1 ],
			blinking: false,
			useCT: (params.ct !== undefined)
		}
	];

	if (lights[0].useCT)
	{
		lights[0].ct = params.ct;
		lights[0].ctbri = params.ctbri;
	}

	sendLightsToHueBridge(lights);
}


function getBaseActivityPath(activityID, activityType)
{
	return activityType + "/" + activityID;
}


function setActivity(activityID, activityType, activity)
{
	if (!checkActivityType(activityType))
		return null;

	if (activityID === null)
	{
		activityID = findFreeActivityID(activityType);
		lInfo.logln("setActivity: using new free activity ID " + activityID + " (type: " + activityType + ")");
	}
	else
	{
		deleteActivity(activityID, activityType);
		activityID = String(activityID);
		lInfo.logln("setActivity: using existing activity ID " + activityID + " (type: " + activityType + ")");
	}

	var basePropertyPath = getBaseActivityPath(activityID, activityType);
	lInfo.logln("setActivity: basePropertyPath: " + basePropertyPath);

	var lightsPath = basePropertyPath + "/lights";
	Property.setProperty(lightsPath, JSON.stringify(activity.lights));
	Property.setFlag(lightsPath, "ARCHIVE", true);

	if (activityType === ActivityTypes.GLOBAL_ACTIVITY)
	{
		var activatedPath = basePropertyPath + "/activated";
		Property.setProperty(activatedPath, activity.activated);
		Property.setFlag(activatedPath, "ARCHIVE", true);
	}

	var triggersPath = basePropertyPath + "/triggers";
	for (var i = 0; i < activity.triggers.length; ++i)
	{
		var trigger = activity.triggers[i];
		var triggerPath = triggersPath + "/" + String(i);

		// In the property tree, actually store bus-zone-scene instead of zone-scene
		// if this actitvity is a preset (improves performance)
		if ((activityType === ActivityTypes.PRESET) && (trigger.type === 'zone-scene'))
			trigger.type = 'bus-zone-scene';

		switch (trigger.type)
		{
			case 'zone-scene':
			case 'bus-zone-scene':
				storeTriggerProperty(triggerPath, "type", trigger.type);
				storeTriggerProperty(triggerPath, "zone", trigger.zone);
				storeTriggerProperty(triggerPath, "group", trigger.group);
				storeTriggerProperty(triggerPath, "scene", trigger.scene);
				storeTriggerProperty(triggerPath, "dsid", (trigger.dsid === undefined) ? -1 : trigger.dsid);
				if (trigger.forced !== undefined)
					storeTriggerProperty(triggerPath, "forced", trigger.forced);
				break;

			case 'undo-zone-scene':
				storeTriggerProperty(triggerPath, "type", trigger.type);
				storeTriggerProperty(triggerPath, "zone", trigger.zone);
				storeTriggerProperty(triggerPath, "group", trigger.group);
				storeTriggerProperty(triggerPath, "scene", trigger.scene);
				storeTriggerProperty(triggerPath, "dsid", (trigger.dsid === undefined) ? -1 : trigger.dsid);
				break;

			case 'device-msg':
				storeTriggerProperty(triggerPath, "type", trigger.type);
				storeTriggerProperty(triggerPath, "dsid", trigger.dsid);
				storeTriggerProperty(triggerPath, "msg", trigger.msg);
				if (trigger.buttonIndex !== undefined)
					storeTriggerProperty(triggerPath, "buttonIndex", trigger.buttonIndex);
				break;

			case 'device-binary-input':
				storeTriggerProperty(triggerPath, "type", trigger.type);
				storeTriggerProperty(triggerPath, "dsid", trigger.dsid);
				storeTriggerProperty(triggerPath, "index", trigger.index);
				storeTriggerProperty(triggerPath, "state", trigger.state);
				break;

			case 'device-scene':
				storeTriggerProperty(triggerPath, "type", trigger.type);
				storeTriggerProperty(triggerPath, "dsid", trigger.dsid);
				storeTriggerProperty(triggerPath, "scene", trigger.scene);
				break;

			case 'custom-event':
				storeTriggerProperty(triggerPath, "type", trigger.type);
				storeTriggerProperty(triggerPath, "event", trigger.event);
				break;

			case 'device-sensor':
				storeTriggerProperty(triggerPath, "type", trigger.type);
				storeTriggerProperty(triggerPath, "dsid", trigger.dsid);
				storeTriggerProperty(triggerPath, "eventid", trigger.eventid);
				break;

			case 'state-change':
				storeTriggerProperty(triggerPath, "type", trigger.type);
				storeTriggerProperty(triggerPath, "name", trigger.name);
				storeTriggerProperty(triggerPath, "state", trigger.state);
				break;

			case 'addon-state-change':
				storeTriggerProperty(triggerPath, "type", trigger.type);
				storeTriggerProperty(triggerPath, "addon-id", trigger.addonId);
				storeTriggerProperty(triggerPath, "name", trigger.name);
				storeTriggerProperty(triggerPath, "state", trigger.state);
				break;

			default:
				lWarn.logln('setActivity: unknown trigger type "' + String(trigger.type) + '"');
				break;
		}
	}

	{
		lInfo.logln("setActivity: checking if other activities use the specified triggers");

		var baseActivityNode = Property.getNode(activityType);
		var activityNodes = baseActivityNode.getChildren();
		for (var i = 0; i < activityNodes.length; ++i)
		{
			var activityNode = activityNodes[i];
			if ((activityNode.getName() === activityID) || (activityNode.getName() === "predefined"))
				continue;

			var activityTriggerParent = activityNode.getChild('triggers');
			var activityTriggerNodes = activityTriggerParent.getChildren();
			var numDeletedTriggers = 0;
			var numTriggersInActivity = activityTriggerNodes.length;

			for (var j = 0; j < activity.triggers.length; ++j)
			{
				var trigger = activity.triggers[j];
				for (var k = 0; k < activityTriggerNodes.length; ++k)
				{
					var activityTriggerNode = activityTriggerNodes[k];
					var activityTrigger = convTriggerNodeToObject(activityTriggerNode);
					if (compareTriggers(trigger, activityTrigger))
					{
						lInfo.logln("setActivity: activity of type " + activityType + " with ID " + activityNode.getName() + " is using trigger " + String(k));

						activityTriggerParent.removeChild(activityTriggerNode);
						numDeletedTriggers++;
					}
				}
			}

			lInfo.logln("setActivity: count for activity of type " + activityType + " with ID " + activityNode.getName() + ": num deleted " + String(numDeletedTriggers) + " num in activity " + String(numTriggersInActivity));

			if (numDeletedTriggers === numTriggersInActivity)
			{
				lInfo.logln("setActivity: activity of type " + activityType + " with ID " + activityNode.getName() + " has no more triggers -> deleting activity");
				baseActivityNode.removeChild(activityNode);
			}
		}
	}

	Property.store();

	registerTrigger(Property.getNode(basePropertyPath).getPath(), "hueController.handleTrigger", { });

	return activityID;
}


function findLightInActivity(activity, lightID)
{
	var lights = activity.lights;
	for (var i = 0; i < lights.length; ++i)
	{
		var light = lights[i];
		if (light.id === lightID)
			return light;
	}

	return null;
}


function compareTriggers(triggerA, triggerB)
{
	// simple trigger comparison

	// investigate values
	// values which exist in B not not in A are not handled,
	// since such a case is not part of the trigger specs
	for (var v in triggerA)
	{
		if (v === 'type')
		{
			// treat (bus-)zone-scene types specially
			// zone-scene and bus-zone-scene are to be treated as equal
			if ((triggerA.type === 'bus-zone-scene') && (triggerB.type === 'zone-scene'))
				continue;
			if ((triggerA.type === 'zone-scene') && (triggerB.type === 'bus-zone-scene'))
				continue;
			else if (triggerA.type !== triggerB.type)
				return false;
		}
		else if (triggerB[v] === undefined)
			continue; // ignore values which are not present in the other trigger
		else if (triggerA[v] !== triggerB[v])
			return false;
	}

	return true;
}


function getStandardLightSettings(activityType, triggers)
{
	var standardActivities = (activityType === ActivityTypes.GLOBAL_ACTIVITY) ? standardGlobalActivities : standardPresets;

	// visit each entry in the standard activity table,
	// looking for a match in the triggers
	for (var i = 0; i < standardActivities.length; ++i)
	{
		var stdActivity = standardActivities[i];

		// compare each trigger from the standard activity
		// to each trigger from the arguments list
		// if a trigger matches, use this activity's light settings
		for (var t1 = 0; t1 < stdActivity.triggers.length; ++t1)
		{
			var stdTrigger = stdActivity.triggers[t1];
			for (var t2 = 0; t2 < triggers.length; ++t2)
			{
				var trigger = triggers[t2];
				if (compareTriggers(stdTrigger, trigger))
					return stdActivity.lightSettings;
			}
		}
	}

	// if this place is reached, no matching trigger was found in the table
	// -> use fallback default
	return standardLightSettings[activityType];
}


function checkLightsOfActivities(activityType)
{
	try
	{
		var baseNode = Property.getNode(activityType);
		if (!baseNode)
			return;

		var nodesToRemove = [];

		var hueLights = null;

		// for global activities, all lights are relevant
		if (activityType === ActivityTypes.GLOBAL_ACTIVITY)
			hueLights = hueQuery("lights");

		var activityNodes = baseNode.getChildren();
		for (var i = 0; i < activityNodes.length; ++i)
		{
			var activityNode = activityNodes[i];

			var idstr = activityNode.getName();
			if (idstr === "predefined")
			{
				lInfo.logln("checkLightsOfActivities: skipping predefined activities");
				continue;
			}

			lInfo.logln("checkLightsOfActivities: checking activity at path " + activityNode.getPath());

			var activity = activityNodeToObject(idstr, activityType, activityNode);

			var lights = activity.lights;
			var lightsModified = false;

			// for presets, only the lights associated with the zone are relevant
			if (activityType === ActivityTypes.PRESET)
			{
				hueLights = getLightsInZoneAsObject(activity.triggers[0].zone);
				lInfo.logln("checkLightsOfActivities: relevant HUE lights: " + JSON.stringify(hueLights));
			}

			// check if any entries in the lights list are stale and need to be removed
			// do so by filtering out lights
			var updatedLights = [];
			for (var lidx = 0; lidx < lights.length; ++lidx)
			{
				var lightID = lights[lidx].id;
				if (hueLights[lightID] !== undefined)
				{
					updatedLights.push(lights[lidx]);
				}
				else
				{
					lInfo.logln("checkLightsOfActivities: detected stale light ID " + String(lightID));
					lightsModified = true;
				}
			}
			lights = updatedLights;

			// check if lights are missing from the list
			var lightsToAdd = [];
			for (var lightID in hueLights)
			{
				var idFound = false;
				for (var lidx = 0; lidx < lights.length; ++lidx)
				{
					if (lights[lidx].id === lightID)
					{
						idFound = true;
						break;
					}
				}

				if (!idFound)
				{
					lInfo.logln("checkLightsOfActivities: HUE bulb with light ID " + String(lightID) + " is not present in lights list");
					lightsToAdd.push(lightID);
				}
			}

			// add necessary lights
			if (lightsToAdd.length > 0)
			{
				lightsModified = true;

				var lightSettings = getStandardLightSettings(activityType, activity.triggers);
				for (var lidx = 0; lidx < lightsToAdd.length; ++lidx)
				{
					// make deep copy of settings
					var newLight = { id: lightsToAdd[lidx] };
					for (var v in lightSettings)
						newLight[v] = lightSettings[v];

					lights.push(newLight);
				}
			}

			// store updated light list
			if (lightsModified)
			{
				if ((activityType === ActivityTypes.PRESET) && (lights.length === 0))
					nodesToRemove.push(activityNode);
				else
				{
					var lightsNode = activityNode.getChild("lights");
					if (!lightsNode)
						lWarn.logln("checkLightsOfActivities: node at path " + activityNode.getPath() + " does not have a \"lights\" subnode");
					else
					{
						lightsNode.setValue(JSON.stringify(lights));
						lightsNode.setFlag('ARCHIVE', true);
					}
				}
			}

			if (lightsModified)
				lInfo.logln("checkLightsOfActivities: lights list for activity at path " + activityNode.getPath() + " had to be updated");
			else
				lInfo.logln("checkLightsOfActivities: lights list for activity at path " + activityNode.getPath() + " is up-to-date");
		}

		for (var i = 0; i < nodesToRemove.length; ++i)
		{
			var node = nodesToRemove[i];
			lInfo.logln("checkLightsOfActivities: preset at path " + node.getPath() + " needs to be removed, since no lights are associated with its zone");
			baseNode.removeChild(node);
		}

		Property.store();
	}
	catch (err)
	{
		lWarn.logln("checkLightsOfActivities: error: " + String(err));
	}
}


function checkLightsOfAllActivities()
{
	checkLightsOfActivities(ActivityTypes.GLOBAL_ACTIVITY);
	checkLightsOfActivities(ActivityTypes.PRESET);
}


function activityNodeToObject(activityID, activityType, node)
{
	var activity = {};

	if (node === null)
	{
		lWarn.logln("activityNodeToObject: node is null");
		return null;
	}

	var lightsNode = node.getChild("lights");
	if (!lightsNode)
	{
		lWarn.logln("activityNodeToObject: node at path " + node.getPath() + " does not have a \"lights\" subnode");
		return null;
	}
	activity.lights = JSON.parse(htmlUnescape(lightsNode.getValue()));

	if (activityType === ActivityTypes.GLOBAL_ACTIVITY)
	{
		var activatedNode = node.getChild("activated");
		if (!activatedNode)
		{
			lWarn.logln("activityNodeToObject: node at path " + node.getPath() + " does not have an \"activated\" subnode");
			return null;
		}
		activity.activated = activatedNode.getValue();
	}

	var triggersNode = node.getChild("triggers");
	if (!triggersNode)
	{
		lWarn.logln("activityNodeToObject: node at path " + node.getPath() + " does not have a \"triggers\" subnode");
		return null;
	}

	activity.triggers = [];
	var triggersChildren = triggersNode.getChildren();
	for (var i = 0; i < triggersChildren.length; ++i)
	{
		var triggerNode = triggersChildren[i];
		var triggerType = triggerNode.getChild("type").getValue();
		var triggerObj = { type: triggerType };

		// In the property tree, zone-scene triggers are actually stored as bus-zone-scene ones
		// for presets, but to the outside, this shall be invisible -> replace "bus-zone-scene"
		// with "zone-scene" then just before returning the list of triggers
		if ((activityType === ActivityTypes.PRESET) && (triggerType === 'bus-zone-scene'))
			triggerType = 'zone-scene';

		switch (triggerType)
		{
			case 'zone-scene':
			case 'bus-zone-scene':
				triggerObj.zone = triggerNode.getChild("zone").getValue();
				triggerObj.group = triggerNode.getChild("group").getValue();
				triggerObj.scene = triggerNode.getChild("scene").getValue();
				triggerObj.dsid = triggerNode.getChild("dsid").getValue();
				if (triggerNode.getChild("forced") !== null)
					triggerObj.forced = triggerNode.getChild("forced").getValue();
				break;

			case 'undo-zone-scene':
				triggerObj.zone = triggerNode.getChild("zone").getValue();
				triggerObj.group = triggerNode.getChild("group").getValue();
				triggerObj.scene = triggerNode.getChild("scene").getValue();
				triggerObj.dsid = triggerNode.getChild("dsid").getValue();
				break;

			case 'device-msg':
				triggerObj.zone = triggerNode.getChild("zone").getValue();
				triggerObj.dsid = triggerNode.getChild("dsid").getValue();
				triggerObj.msg = triggerNode.getChild("msg").getValue();
				if (triggerNode.getChild("buttonIndex") !== null)
					triggerObj.zone = triggerNode.getChild("buttonIndex").getValue();
				break;

			case 'device-binary-input':
				triggerObj.dsid = triggerNode.getChild("dsid").getValue();
				triggerObj.index = triggerNode.getChild("index").getValue();
				triggerObj.state = triggerNode.getChild("state").getValue();
				break;

			case 'device-scene':
				triggerObj.dsid = triggerNode.getChild("dsid").getValue();
				triggerObj.scene = triggerNode.getChild("scene").getValue();
				break;

			case 'custom-event':
				triggerObj.event = triggerNode.getChild("event").getValue();
				break;

			case 'device-sensor':
				triggerObj.dsid = triggerNode.getChild("dsid").getValue();
				triggerObj.eventid = triggerNode.getChild("eventid").getValue();
				break;

			case 'state-change':
				triggerObj.name = triggerNode.getChild("name").getValue();
				triggerObj.state = triggerNode.getChild("state").getValue();
				break;

			case 'addon-state-change':
				triggerObj.addonId = triggerNode.getChild("addon-id").getValue();
				triggerObj.name = triggerNode.getChild("name").getValue();
				triggerObj.state = triggerNode.getChild("state").getValue();
				break;

			default:
				break;
		}

		activity.triggers.push(triggerObj);
	}

	return activity;
}


function getActivity(activityID, activityType)
{
	if (!checkActivityType(activityType))
		return null;

	var baseNode = Property.getNode(activityType);
	if (baseNode)
	{
		var idstr = String(activityID);
		var child = baseNode.getChild(idstr);
		if (child)
		{
			lInfo.logln('getActivity: found activity of type "' + activityType + '" with ID ' + idstr);
			return activityNodeToObject(activityID, activityType, child);
		}
		else
		{
			lInfo.logln('getActivity: found no activity of type "' + activityType + '" with ID ' + idstr);
			return null;
		}
	}
	else
	{
		lInfo.logln('getActivity: found no subnode for activities of type "' + activityType + '"');
		return null;
	}
}


function getAllActivities(activityType)
{
	if (!checkActivityType(activityType))
		return null;

	lInfo.logln("getAllActivities: returning all activities of type " + activityType);

	var baseNode = Property.getNode(activityType);
	var activities = {};

	if (baseNode)
	{
		var children = baseNode.getChildren();
		for (var i = 0; i < children.length; ++i)
		{
			var child = children[i];
			var idstr = child.getName();
			if (idstr === "predefined")
			{
				lInfo.logln("getAllActivities: skipping predefined triggers");
				continue;
			}
			var activity = activityNodeToObject(idstr, activityType, child);
			if (activity !== null)
				activities[idstr] = activity;
		}
	}

	return activities;
}


function convTriggerNodeToObject(triggerNode)
{
	var triggerType = triggerNode.getChild("type").getValue();
	var triggerObj = { type: triggerType };

	switch (triggerType)
	{
		case 'zone-scene':
		case 'bus-zone-scene':
			triggerObj.zone = triggerNode.getChild("zone").getValue();
			triggerObj.group = triggerNode.getChild("group").getValue();
			triggerObj.scene = triggerNode.getChild("scene").getValue();
			triggerObj.dsid = triggerNode.getChild("dsid").getValue();
			if (triggerNode.getChild("forced") !== null)
				triggerObj.forced = triggerNode.getChild("forced").getValue();
			break;

		case 'undo-zone-scene':
			triggerObj.zone = triggerNode.getChild("zone").getValue();
			triggerObj.group = triggerNode.getChild("group").getValue();
			triggerObj.scene = triggerNode.getChild("scene").getValue();
			triggerObj.dsid = triggerNode.getChild("dsid").getValue();
			break;

		case 'device-msg':
			triggerObj.zone = triggerNode.getChild("zone").getValue();
			triggerObj.dsid = triggerNode.getChild("dsid").getValue();
			triggerObj.msg = triggerNode.getChild("msg").getValue();
			if (triggerNode.getChild("buttonIndex") !== null)
				triggerObj.zone = triggerNode.getChild("buttonIndex").getValue();
			break;

		case 'device-binary-input':
			triggerObj.dsid = triggerNode.getChild("dsid").getValue();
			triggerObj.index = triggerNode.getChild("index").getValue();
			triggerObj.state = triggerNode.getChild("state").getValue();
			break;

		case 'device-scene':
			triggerObj.dsid = triggerNode.getChild("dsid").getValue();
			triggerObj.scene = triggerNode.getChild("scene").getValue();
			break;

		case 'custom-event':
			triggerObj.event = triggerNode.getChild("event").getValue();
			break;

		case 'device-sensor':
			triggerObj.dsid = triggerNode.getChild("dsid").getValue();
			triggerObj.eventid = triggerNode.getChild("eventid").getValue();
			break;

		case 'state-change':
			triggerObj.name = triggerNode.getChild("name").getValue();
			triggerObj.state = triggerNode.getChild("state").getValue();
			break;

		case 'addon-state-change':
			triggerObj.addonId = triggerNode.getChild("addon-id").getValue();
			triggerObj.name = triggerNode.getChild("name").getValue();
			triggerObj.state = triggerNode.getChild("state").getValue();
			break;

		default:
			break;
	}

	return triggerObj;
}


function filterAlreadyUsedTriggers(activityType, triggers)
{
	// deep copy of array
	var filteredTriggers = [];
	for (var i = 0; i < triggers.length; ++i)
		filteredTriggers[i] = triggers[i];

	lInfo.logln("filterAlreadyUsedTriggers: filtering if necessary");

	if (!checkActivityType(activityType))
		return null;

	var baseNode = Property.getNode(activityType);
	var activities = {};

	if (baseNode)
	{
		var activityNodes = baseNode.getChildren();
		if (activityNodes)
		{
			// visit each activity node
			for (var i = 0; i < activityNodes.length; ++i)
			{
				var activityNode = activityNodes[i];

				var actChildNodes = activityNode.getChildren();
				var triggerParentNode = null;
				for (var j = 0; j < actChildNodes.length; ++j)
				{
					if (actChildNodes[j].getName() === 'triggers')
					{
						triggerParentNode = actChildNodes[j];
						break;
					}
				}

				if (!triggerParentNode)
					continue;

				var triggerNodes = triggerParentNode.getChildren();
				if (!triggerNodes)
					continue;

				// visit each trigger node
				for (var j = 0; j < triggerNodes.length; ++j)
				{
					var triggerNode = triggerNodes[j];

					var triggerObj = convTriggerNodeToObject(triggerNode);

					// compare the triggers in filteredTriggers with the trigger from the node
					// if they are a match, remove the entry from filteredTriggers
					for (var k = 0; k < filteredTriggers.length; ++k)
					{
						if (compareTriggers(filteredTriggers[k], triggerObj))
						{
							lInfo.logln("filterAlreadyUsedTriggers: filtered trigger " + JSON.stringify(filteredTriggers[k]));
							filteredTriggers.splice(k, 1);
							break;
						}
					}
				}
			}
		}
	}

	if (filteredTriggers.length === triggers.length)
		lInfo.logln("filterAlreadyUsedTriggers: no triggers were filtered");

	return filteredTriggers;
}


function createStandardGlobalActivities()
{
	try
	{
		lInfo.logln("createStandardGlobalActivities: adding standard global activities");

		var foundLights = hueQuery("lights");
		for (var actidx = 0; actidx < standardGlobalActivities.length; ++actidx)
		{
			var standardActivity = standardGlobalActivities[actidx];
			var lightSettings = standardActivity.lightSettings;

			// check existing global activities, and if some of the triggers
			// specified in the triggers are already used, remove them from
			// the triggers array
			// if all triggers are used somewhere else (-> the array becomes empty),
			// do not add the activity
			triggers = filterAlreadyUsedTriggers(ActivityTypes.GLOBAL_ACTIVITY, standardActivity.triggers);
			if (triggers.length === 0)
			{
				lInfo.logln("createStandardGlobalActivities: all triggers for standard activity " + String(actidx) + " are used elsewhere -> not adding activity");
				continue;
			}

			var newActivity =
			{
				triggers: triggers,
				lights: [],
				activated: true
			};

			// do deep copies of the light settings and add the light IDs to them
			for (var lidx in foundLights)
			{
				var newLight = { id: String(lidx) };
				for (var v in lightSettings)
					newLight[v] = lightSettings[v];
				newActivity.lights.push(newLight);
			}

			setActivity(null, ActivityTypes.GLOBAL_ACTIVITY, newActivity);
		}
	}
	catch (err)
	{
		lWarn.logln("createStandardGlobalActivities: error while creating standard global activities: " + String(err));
	}
}

function createStandardPresetsForAllZones()
{
	var lightsZoneAssoc = getHueLightZoneAssotiations();
	for (var i = 0; i < lightsZoneAssoc.length; ++i)
	{
		createStandardPresets(lightsZoneAssoc[i].zoneID);
	}
}


function createStandardPresets(zoneID)
{
	try
	{
		lInfo.logln("createStandardPresets: adding standard presets for zone with ID " + String(zoneID));

		var foundLights = getLightsInZone(zoneID);
		for (var actidx = 0; actidx < standardPresets.length; ++actidx)
		{
			var standardActivity = standardPresets[actidx];
			var lightSettings = standardActivity.lightSettings;

			var newActivity =
			{
				triggers: [],
				lights: [],
				activated: true
			};

			// do deep copies of all triggers and add the zone ID to them
			for (var tidx = 0; tidx < standardActivity.triggers.length; ++tidx)
			{
				var trigger = standardActivity.triggers[tidx];
				var newTrigger = { zone: zoneID };
				for (var v in trigger)
					newTrigger[v] = trigger[v];
				newActivity.triggers.push(newTrigger);
			}

			// check existing presets, and if some of the triggers
			// specified in the triggers are already used, remove them from
			// the triggers array
			// if all triggers are used somewhere else (-> the array becomes empty),
			// do not add the activity
			newActivity.triggers = filterAlreadyUsedTriggers(ActivityTypes.PRESET, newActivity.triggers);
			if (newActivity.triggers.length === 0)
			{
				lInfo.logln("createStandardPresets: all triggers for standard preset " + String(actidx) + " are used elsewhere -> not adding preset");
				continue;
			}

			// do deep copies of the light settings and add the light IDs to them
			for (var lidx = 0; lidx < foundLights.length; ++lidx)
			{
				var newLight = { id: String(foundLights[lidx]) };
				for (var v in lightSettings)
					newLight[v] = lightSettings[v];
				newActivity.lights.push(newLight);
			}

			setActivity(null, ActivityTypes.PRESET, newActivity);
		}
	}
	catch (err)
	{
		lWarn.logln("createStandardPresets: error while creating standard presets for zone with ID " + String(zoneID) + ": " + String(err));
	}
}

